/* We include this copyright since we're using Cacti for some stuff */

/*------------------------------------------------------------
 *  Copyright 1994 Digital Equipment Corporation and Steve Wilton
 *                         All Rights Reserved
 *
 * Permission to use, copy, and modify this software and its documentation is
 * hereby granted only under the following terms and conditions.  Both the
 * above copyright notice and this permission notice must appear in all copies
 * of the software, derivative works or modified versions, and any portions
 * thereof, and both notices must appear in supporting documentation.
 *
 * Users of this software agree to the terms and conditions set forth herein,
 * and hereby grant back to Digital a non-exclusive, unrestricted, royalty-
 * free right and license under any changes, enhancements or extensions
 * made to the core functions of the software, including but not limited to
 * those affording compatibility with other hardware or software
 * environments, but excluding applications which incorporate this software.
 * Users further agree to use their best efforts to return to Digital any
 * such changes, enhancements or extensions that they make and inform Digital
 * of noteworthy uses of this software.  Correspondence should be provided
 * to Digital at:
 *
 *                       Director of Licensing
 *                       Western Research Laboratory
 *                       Digital Equipment Corporation
 *                       100 Hamilton Avenue
 *                       Palo Alto, California  94301
 *
 * This software may be distributed (but not offered for sale or transferred
 * for compensation) to third parties, provided such third parties agree to
 * abide by the terms and conditions of this notice.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND DIGITAL EQUIPMENT CORP. DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL DIGITAL EQUIPMENT
 * CORPORATION BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>

#include "resource.h"
#include "machine.h"
#include "cache.h"
#include "power.h"
#include "def.h"
#include "areadef.h"
#include "flp.h"
#include "RC.h"

void calculate_time(result_type *result, arearesult_type *arearesult,
                    area_type *arearesult_subbanked, parameter_type *parameters, double *NSubbanks);

// hotspot-related
flp_t *flp; 
double power_vector[TOTAL_MODULES + MODULE_L2_EXTRA + EXTRA]; // +2 because L2 include 3 parts: left, right and bottom
double temperature_vector[TOTAL_MODULES + MODULE_L2_EXTRA + EXTRA]; // +2 because L2 include 3 parts: left, right and bottom

// SimpleScalar-related
int opcode_length = 32;
int inst_length = 64;

extern long fast_forward_for_stable;

extern tick_t sim_cycle;
extern counter_t sim_total_insn;
extern int ruu_ifq_size;
extern int ruu_decode_width;
extern int ruu_issue_width;
extern int ruu_commit_width;
extern int RUU_size;
extern int LSQ_size;
//extern int data_width;
extern int res_ialu;
extern int res_fpalu;
extern int res_memport;
//int nvreg_width;
//int npreg_width;
extern int bimod_config[];
extern struct cache_t *cache_dl1;
extern struct cache_t *cache_il1;
extern struct cache_t *cache_dl2;
extern struct cache_t *dtlb;
extern struct cache_t *itlb;
extern struct res_pool *fu_pool;

/* 2-level predictor config (<l1size> <l2size> <hist_size> <xor>) */
extern int twolev_config[];
  
/* combining predictor config (<meta_table_size> */
extern int comb_config[];
    
/* return address stack (RAS) size */
extern int ras_size;
  
/* BTB predictor config (<num_sets> <associativity>) */
extern int btb_config[];

char module_name[TOTAL_MODULES][32] = {
  "decode", "branch", "rename", "ruu", "lsq", "ialu1", "ialu2", "ialu3", "ialu4", "falu1", "falu2",
  "int_reg", "fp_reg", "itlb", "il1", "dtlb", "dl1", "l2"
};

/* for debug purpose */
FILE *debugFP;
char file_config[256] = "power_conf";
char file_output[256];

//double maxTemp = 100;
//unsigned INTERVAL = 1000;
unsigned last_cycle = 0;

// default setting voltage scaling setting:
double vDD = 1.0; // default, for 3GHz frequency
double freq = 3.0e9;
double t_max = 110; // max temperature constraint in C
double max_gap = 30; // max temperature gap constraint in C
unsigned long max_violate, max_gap_violate;

struct MEM_CONFIG mem_conf[TOTAL_MODULES]; // only used for memory-based units, not used by memory-less units 
struct COMPONENT module[TOTAL_MODULES];
struct FORMULA sram;

double total_power; /* whole system power without any throttling */
double last_total_power_clock; /* whole system power with only clock gating in last cycle */
double last_total_power_gating; /* whole system power with power gating in last cycle */
double max_didt_power;   /* max didt power without any throttling */
double max_didt_power_clock;   /* max didt power with only clock gating */
double max_didt_power_gating;   /* max didt power with power gating */

/* overall temperature */ 
double total_tmpr = T_INIT; 
double total_lastTmprPower = 0.0;

void reset_counters()
{
  int i;

  for (i = 0; i < TOTAL_MODULES; i ++) {
    module[i].access = 0;
  }
}

/* initialize power statstics at the begining */
void init_module(COMPONENT *c)
{
  c->totalPower = 0.0;
  c->totalPowerClock = 0.0;
  c->totalPowerGating = 0.0;
  c->lastPowerClock = 0.0;
  c->lastPowerGating = 0.0;
  c->presPowerClock = 0.0;
  c->presPowerGating = 0.0;
  c->totalDynamicPower = 0.0;
  c->totalStandbyPower = 0.0;
  c->temperature = T_INIT;
  c->lastTmprPower = 0.0;

  c->tauHeat = 0.0001;
  c->tauCool = 0.0001;
}

void init_all_modules()
{
  int i;
  
  for (i = 0; i < TOTAL_MODULES; i ++) {
    init_module(&(module[i])); 
  }
}


/* clear frequency, temperature and FORMULA */
void para_clear()
{
  sram.pd_a = 0.0;
  sram.pd_b = 0.0;
  sram.ps_a = 0.0;
  sram.ps_b = 0.0;
  sram.ps_c = 0.0;
  sram.ps_d = 0.0;
  sram.ps_x = 0.0;
  sram.ps_y = 0.0;
  sram.ps_z = 0.0;
  sram.pi_a = 0.0;
  sram.pi_b = 0.0;
  sram.pi_c = 0.0;
  sram.pi_d = 0.0;
  sram.pi_x = 0.0;
  sram.pi_y = 0.0;
  sram.pi_z = 0.0;
}


int getName(char *str)
{
  int i;

  for (i = 0; i < TOTAL_MODULES; i ++) {
    if (!strcmp(str, module_name[i]))
      return i;
  }

  return -1;
}	 

void ftoa(char str[], float f)
{
  double a, b;
  int dig;
  int i;

  a = floor(f);
  b = floor((f - a) * 10); // only consider one digital after point  
  i = 0;

  while (a > 0) {
    dig = (int)(a - floor(a / 10) * 10);
    str[i ++] = dig + 48;
    a = floor(a / 10);
  }
  str[i ++] = '.';
  str[i ++] = (int)b + 48;
  str[i ++] = 0;
  strcat(str, "GHz");
  
  return;
}

#ifdef _HOTSPOT_
void power_tmpr_vector(double pwr_vector[], double tmpr_vector[])
{
  int i;

  for (i = 0; i < TOTAL_MODULES - 1; i ++) { // L2 cache should be treated seperately
    tmpr_vector[i] = module[i].temperature + 273.15;
    pwr_vector[i] = module[i].lastTmprPower / (INTERVAL / freq) * U;
  }
  
  // handle L2 cache
  tmpr_vector[MODULE_L2_LEFT] = module[MODULE_L2].temperature + 273.15;
  tmpr_vector[MODULE_L2_BOTTOM] = module[MODULE_L2].temperature + 273.15;
  tmpr_vector[MODULE_L2_RIGHT] = module[MODULE_L2].temperature + 273.15;

  pwr_vector[MODULE_L2_LEFT] = module[MODULE_L2].lastTmprPower * U * freq / INTERVAL * 1.0 / 7.0;
  pwr_vector[MODULE_L2_BOTTOM] = module[MODULE_L2].lastTmprPower * U * freq / INTERVAL  * 5.0 / 7.0;
  pwr_vector[MODULE_L2_RIGHT] = module[MODULE_L2].lastTmprPower * U * freq / INTERVAL * 1.0 / 7.0;
}
#endif

void power_init()
{
  FILE *fp;
  int i, id, omit_lateral;
  char freq_suf[255];
  char str[256];

  #ifdef _DEBUG_
  if ((debugFP = fopen("debug.out", "w")) == NULL) {
    printf("can not open debug.out to write in debug mode!\n");
    exit(1);
  }
  #endif

  max_violate = 0;
  max_gap_violate = 0;

  #ifdef _HOTSPOT_
  // hotspot
  flp = read_flp("ev6.flp");
  create_RC_matrices(flp, omit_lateral);
  for (i = 0; i < EXTRA; i ++) { // initialize temperature for heat sink
    temperature_vector[TOTAL_MODULES + MODULE_L2_EXTRA + i] = T_INIT + 273.15;
  }
  #endif

  //ftoa(freq_suf, freq / 1e9);
  //strcat(file_config, freq_suf);
  //strcat(file_output, freq_suf);
  //printf("file_config is %s and file_output is %s\n", file_config, file_output);
  // file_output is temperorialy unused.

  for (i = 0; i < TOTAL_MODULES; i ++) {
    module[i].total_access = 0; // rest_counters() will be called in every cycles, so reset total_access should not be included in reset_counters.
  }
  reset_counters();
  init_all_modules();

  if ((fp = fopen(file_config, "r")) == NULL) {
    printf("Can not open configuration file %s\n", file_config);
    exit(1);
  }

  // get area information from configuration file
  while(fscanf(fp, "%s", str) != EOF) {
    if (!strcmp(str, "<end>"))
      break;
    else if (!strcmp(str, "<begin>"))
      continue;
    else if (!strcmp(str, "Pd")) {
      fscanf(fp, "%lf %lf", &(sram.pd_a), &(sram.pd_b));
    } else if (!strcmp(str, "Ps")) {
      fscanf(fp, "%lf %lf %lf %lf %lf %lf %lf", &(sram.ps_x), &(sram.ps_y), &(sram.ps_a), &(sram.ps_b), &(sram.ps_z), &(sram.ps_c), &(sram.ps_d));
    } else if (!strcmp(str, "Pi")) {
      fscanf(fp, "%lf %lf %lf %lf %lf %lf %lf", &(sram.pi_x), &(sram.pi_y), &(sram.pi_a), &(sram.pi_b), &(sram.pi_z), &(sram.pi_c), &(sram.pi_d));
    } else if (!strcmp(str, "fmax")) {
      fscanf(fp, "%lf", &freq);
    } else if (!strcmp(str, "vdd")) {
      fscanf(fp, "%lf", &vDD);
    } else if (!strcmp(str, "t_max")) {
      fscanf(fp, "%lf", &t_max);
    } else if (!strcmp(str, "max_gap")) {
      fscanf(fp, "%lf", &max_gap);
    } else {
      if ((id = getName(str)) >= 0) { // valid id for module
        fscanf(fp, "%lf", &(module[id].totalArea));
      }
    }
  }

  fclose(fp);

  calculate_module_power();
}

// called at any time to see the power under temperature at that moment
void power_output(FILE *fp)
{
  int i;

  fprintf(fp, "frequency is %e\n", freq);
  fprintf(fp, "supply voltage is %lf\n", vDD);
  fprintf(fp, "Cycle %u\n", sim_cycle);
  fprintf(fp, "Average power is in uJ/cycle, cumulated power is in uJ\n");
  for (i = 0; i < TOTAL_MODULES; i ++) {
    fprintf(fp, "For %s,\n  \
		active power is \t%18.14lf uJ/cycle (%18.14lf Watt),\n  \
		dynamic power is \t%18.14lf uJ/cycle (%18.14lf Watt),\n  \
		standby power is \t%18.14lf uJ/cycle (%18.14lf Watt),\n  \
		inactive power is \t%18.14lf uJ/cycle (%18.14lf Watt),\n  \
		area is \t%18.14lf mm^2, thermal resistance is \t%18.14lf\n", 
  		module_name[i], module[i].activePower, module[i].activePower * freq * U, 
		module[i].dynPower, module[i].dynPower * freq * U,
		module[i].standbyPower/freq/U, module[i].standbyPower,
		module[i].inactivePower/freq/U, module[i].inactivePower,
		module[i].area * 1e6, module[i].thmResis);
  }
}

// output to file_output
void collected_power_output(FILE *fp)
{
  double total_power, total_power_clock, total_power_gating;
  double total_dynamic_power = 0.0, total_standby_power = 0.0;

  int i;
  
  power_output(fp);
  for (i = 0; i < TOTAL_MODULES; i ++) {
    total_power += module[i].totalPower;
    total_power_clock += module[i].totalPowerClock; 
    total_power_gating += module[i].totalPowerGating; 
    total_dynamic_power += module[i].totalDynamicPower;
    total_standby_power += module[i].totalStandbyPower;
  }  

  fprintf(fp, "\ntotal power without gating: %lf uJ\n", total_power);
  fprintf(fp, "average power per cycle without gating: %lf uJ/cycle (%lf Watt)\n", 
  		total_power / sim_cycle, total_power / sim_cycle * U * freq);
  fprintf(fp, "average power per instruction without gating: %lf uJ / cycle (%lf Watt)\n", 
  		total_power / sim_total_insn, total_power / sim_total_insn * U * freq);
  for (i = 0; i < TOTAL_MODULES; i ++) { 
    fprintf(fp, "total %s power without gating: %lf\n", module_name[i], module[i].totalPower);
  }

  fprintf(fp, "\ntotal power with only clock gating: %lf uJ\n", total_power_clock);
  fprintf(fp, "average power per cycle with only clock gating: %lf uJ/cycle (%lf Watt)\n", 
  		total_power_clock / sim_cycle, total_power_clock / sim_cycle * U * freq);
  fprintf(fp, "average power per instruction with only clock gating: %lf uJ/cycle (%lf Watt)\n", 
  		total_power_clock / sim_total_insn, total_power_clock / sim_total_insn * U * freq);
  for (i = 0; i < TOTAL_MODULES; i ++) { 
    fprintf(fp, "total %s power with only clock gating: %lf\n", module_name[i], module[i].totalPowerClock);
  }

  fprintf(fp, "\ntotal power with power gating: %lf uJ\n", total_power_gating);
  fprintf(fp, "average power per cycle with power gating: %lf uJ/cycle (%lf Watt)\n", 
  		total_power_gating / sim_cycle, total_power_gating / sim_cycle * freq * U);
  fprintf(fp, "average power per instruction with power gating: %lf uJ/cycle (%lf Watt)\n", 
  		total_power_gating / sim_total_insn, total_power_gating / sim_total_insn * U * freq);
  for (i = 0; i < TOTAL_MODULES; i ++) { 
    fprintf(fp, "total %s power with both clock and power gating (utilization rate): %lf\n", module_name[i], module[i].totalPowerGating);
  }

  fprintf(fp, "\ntotal dynamic power with clock gating: %lf\n", total_dynamic_power);
  fprintf(fp, "average power per cycle with clock gating: %lf uJ/cycle (%lf Watt)\n", 
  		total_dynamic_power / sim_cycle, total_dynamic_power / sim_cycle * freq * U);
  fprintf(fp, "average power per instruction on-path with clock gating: %lf uJ/cycle (%lf Watt)\n", 
  		total_dynamic_power / sim_total_insn, total_dynamic_power / sim_total_insn * U * freq);
  for (i = 0; i < TOTAL_MODULES; i ++) { 
    fprintf(fp, "total %s dynamic power with clock gating: %lf\n", module_name[i], module[i].totalDynamicPower);
  }

  fprintf(fp, "\ntotal standby power without gating: %lf\n", total_standby_power);
  fprintf(fp, "average power per cycle with power gating: %lf uJ/cycle (%lf Watt)\n", 
  		total_standby_power / sim_cycle, total_standby_power / sim_cycle * U * freq);
  fprintf(fp, "average power per instruction on-path with power gating: %lf uJ/cycle (%lf Watt)\n", 
  		total_standby_power / sim_total_insn, total_standby_power / sim_total_insn * U * freq);
  for (i = 0; i < TOTAL_MODULES; i ++) { 
    fprintf(fp, "total %s standby power without gating: %lf\n", module_name[i], module[i].totalStandbyPower);
  }

  for (i = 0; i < TOTAL_MODULES; i ++) { 
    fprintf(fp, "final %s temperature: %18.14lf\n", module_name[i], module[i].temperature);
  }

  fprintf(fp, "\nmax temperature violation %u\nmax temperature gap violation %u\n", max_violate, max_gap_violate); 
}


/* update power statstics every cycle, here the power statstics are actully energy! */
int update_power() 
{
  int i, class;
  double pres_total_power; // total power w/o gating, to calculate whole system temperature when no gating
  double pres_total_power_clock, last_total_power_clock, pres_total_power_gating, last_total_power_gating;
  double max, min;

  /* note: totalPower, totalPowerClock/Gating are actually energy in uJ */

  // count power for functional units
  class = 1; // class 1 for int add
  for (i = 0; i < 3; i ++) {
    assert(fu_pool->table[class][i] != NULL);
    if (fu_pool->table[class][i]->master->busy) {
      switch (i) {
      case 0: // IALU1
        module[MODULE_IALU1].access = 1;
        break;
      case 1: // IALU2
        module[MODULE_IALU2].access = 1;
        break;
      case 2: // IALU3
        module[MODULE_IALU3].access = 1;
        break;
      default: break;
      }
    }
  }
  for (class = 2; class < 4; class ++) {
    assert(fu_pool->table[class][0] != NULL);
    if (fu_pool->table[class][0]->master->busy) {
      module[MODULE_IALU4].access = 1;
    }
  }
  /* class 4 - 6 for FP add */
  for (class = 4; class < 7; class ++) {
    assert(fu_pool->table[class][0] != NULL);
    if (fu_pool->table[class][0]->master->busy) {
        module[MODULE_FALU1].access = 1;
    }
  }
  /* class 7 - 9 for FP multi/divide */
  for (class = 7; class < 10; class ++) {
    assert(fu_pool->table[class][0] != NULL);
    if (fu_pool->table[class][0]->master->busy) {
        module[MODULE_FALU2].access = 1;
    }
  }

  pres_total_power = 0.0; 
  pres_total_power_clock = 0.0;
  pres_total_power_gating = 0.0;
  for (i = 0; i < TOTAL_MODULES; i ++) {
    module[i].totalPower += module[i].dynPower * (module[i].access == 0 ? 1 : module[i].access) + module[i].standbyPower / freq / U;
    module[i].presTotalPower = module[i].dynPower * (module[i].access == 0 ? 1 : module[i].access) + module[i].standbyPower / freq / U;
    pres_total_power += module[i].dynPower * (module[i].access == 0 ? 1 : module[i].access) + module[i].standbyPower / freq / U;
    module[i].totalStandbyPower += module[i].standbyPower / freq / U;
    if (module[i].access) {  // this requires one counter per module
      module[i].totalPowerClock += module[i].dynPower * module[i].access + module[i].standbyPower / freq / U;
      module[i].totalPowerGating += module[i].dynPower * module[i].access + module[i].standbyPower / freq / U;
      module[i].presPowerClock = module[i].dynPower * module[i].access + module[i].standbyPower / freq / U;
      module[i].presPowerGating = module[i].dynPower * module[i].access + module[i].standbyPower / freq / U;
      module[i].totalDynamicPower += module[i].dynPower * module[i].access;
    } else {
      module[i].totalPowerClock += module[i].standbyPower / freq / U + module[i].dynPower * (1.0 - CLOCK_GATING_REDUCTION);
      module[i].totalPowerGating += module[i].inactivePower / freq / U + module[i].dynPower * (1.0 - CLOCK_GATING_REDUCTION);
      module[i].presPowerClock = module[i].standbyPower / freq / U + module[i].dynPower * (1.0 - CLOCK_GATING_REDUCTION);
      module[i].presPowerGating = module[i].inactivePower / freq / U + module[i].dynPower * (1.0 - CLOCK_GATING_REDUCTION);
      module[i].totalDynamicPower += module[i].dynPower * (1.0 - CLOCK_GATING_REDUCTION);
    }
    pres_total_power_clock += module[i].presPowerClock; 
    pres_total_power_gating += module[i].presPowerGating; 
  }

  /* record the max didt power, neglect the first cycle */  
  if ((sim_cycle > 1) && (fabs(pres_total_power_clock - last_total_power_clock) > fabs(max_didt_power_clock))) 
    max_didt_power_clock = fabs(pres_total_power_clock - last_total_power_clock);
  if ((sim_cycle > 1) && (fabs(pres_total_power_gating - last_total_power_gating) > fabs(max_didt_power_gating))) 
    max_didt_power_gating = fabs(pres_total_power_gating - last_total_power_gating);

  /* record pres_total_power to last_total_power for next cycle */
  last_total_power_clock = pres_total_power_clock;
  last_total_power_gating = pres_total_power_gating;
  
  #ifndef _CLOCKGATING_
  for (i = 0; i < TOTAL_MODULES; i ++) {
    module[i].lastTmprPower += module[i].presTotalPower;
  }
  /* for whole chip temperature */
  total_lastTmprPower += pres_total_power;
  #else 
  for (i = 0; i < TOTAL_MODULES; i ++) {
    module[i].lastTmprPower += module[i].presPowerClock;
  }
  /* for whole chip temperature */
  total_lastTmprPower += pres_total_power_clock;
  #endif
  
  #ifdef _HOTSPOT_
  if (sim_cycle - last_cycle == INTERVAL) {
      last_cycle = sim_cycle;
      power_tmpr_vector(power_vector, temperature_vector);
      compute_temp(power_vector, temperature_vector, flp->n_units, INTERVAL / freq);

      for (i = 0; i < TOTAL_MODULES - 1; i ++) {
        module[i].temperature = temperature_vector[i] - 273.15;
      }
      module[MODULE_L2].temperature = (temperature_vector[MODULE_L2_LEFT] + temperature_vector[MODULE_L2_BOTTOM]
				   + temperature_vector[MODULE_L2_RIGHT]) / 3.0 - 273.15; 

      update_module_power();	

      // reset module lastTmprPower 
      for (i = 0; i < TOTAL_MODULES; i ++) {
        module[i].lastTmprPower = 0.0;
      }

      // thermal constratints
      max = module[0].temperature;
      min = module[0].temperature;
      for (i = 1; i < TOTAL_MODULES - 1; i ++) {
        if (module[i].temperature > max) {
	  max = module[i].temperature;
	} else if (module[i].temperature < min) { 
	  min = module[i].temperature;
	}
      }
      if (max > t_max) {
	max_violate ++;
      } 
      if (max - min > max_gap) {
        max_gap_violate ++;
      }
  }
  #endif

  /* reset counters */
  reset_counters();

  return 1;
}

/******************* power formulas ****************/
double dyn_power_array(int row, int col, double vdd)
{
  return (sram.pd_a * row + sram.pd_b * col) * vdd * vdd / (STD_VDD * STD_VDD) * CACHE_POWER_FACTOR;
}

double std_power_array(int row, int col, double tmpr, double vdd)
{
  double std_power_circuits, std_power_cells;
  
  std_power_circuits = vdd * (sram.ps_x * row + sram.ps_y * col) * (tmpr + 273) * (tmpr + 273)
		       * exp(-(sram.ps_a * vdd + sram.ps_b) / (tmpr + 273));
  std_power_cells = vdd * (sram.ps_z * row * col) * (tmpr + 273) * (tmpr + 273)
		       * exp(-(sram.ps_c * vdd + sram.ps_d) / (tmpr + 273));

  return (std_power_circuits + std_power_cells) * CACHE_POWER_FACTOR; 
}

double inact_power_array(int row, int col, double tmpr, double vdd)
{
  double inact_power_circuits, inact_power_cells;

  inact_power_circuits = vdd * (sram.pi_x * row + sram.pi_y * col) * (tmpr + 273) * (tmpr + 273)
		       * exp(-(sram.pi_a * vdd + sram.pi_b) / (tmpr + 273));
  inact_power_cells = vdd * (sram.pi_z * row * col) * (tmpr + 273) * (tmpr + 273)
		       * exp(-(sram.pi_c * vdd + sram.pi_d) / (tmpr + 273));

  return (inact_power_circuits + inact_power_cells) * CACHE_POWER_FACTOR;
}

double dynamic_power_comb(int gate_count, double sf, double c_gate, double vdd)
{
  double gate_power, interconnect_power;

  gate_power = (0.5 * (double)gate_count * sf * c_gate * vdd * vdd) / U;
  interconnect_power = 0.5 * sf * rents(gate_count) * vdd * vdd / U;
  return (gate_power + interconnect_power);
}

double standby_power_comb(int gate_count, double tmpr, double vdd)
{
  return gate_count * vdd * (2.48325558 + 2.26321378 + 2.6529006) / 3.0 * 1e-8 * (tmpr + 273) * (tmpr + 273) 
	  * exp(-(sram.ps_a * vdd + sram.ps_b) / (tmpr + 273));
}

// rent's rule for total wire cap
#define LOCAL_INTERMEDIATE_BOUNDARY 21 // gate pitch
#define CW_LOCAL 176.188e-12 
#define CW_INT 180.068e-12 // intermediate wires 
#define ALPHA 1.0 
#define K 4.0
#define P 0.55

double get_tau(double N)
{
  double a, b, c;

  if (P == 0.5) {
    a = 4.0 * N - 4.0 * sqrt(N);
    b = sqrt(N) * (-2.0 * log(N) - 6.0 + 2.0 * log(4.0)) + 4.0 * N - 2.0 / 3.0;
    return (a / b);
  } else {
    a = 2.0 * N * (1.0 - pow(N, P - 1));
    b = - pow(N, P) * (1.0 + 2.0 * P - pow(2.0, 2.0 * P - 1.0)) / (P * (2.0 * P - 1) * (P - 2) * (2 * P - 3));
    c = - 1.0 / (6.0 * P) + 2.0 * sqrt(N) / (2.0 * P - 1.0) - N / (P - 1.0);
    return (a / (b + c));
  }
}

double get_idf(double l, double N)
{
  double tau = get_tau(N);
  double a, b, c;

  if (l < sqrt(N)) {
    a = ALPHA * K / 2.0;
    b = tau * (l * l * l / 3.0 - 2.0 * sqrt(N) * l * l + 2 * N * l);
    c = pow(l, 2.0 * P - 4);
    return (a * b * c);
  } else {
    a = ALPHA * K / 6.0;
    b = tau * (2.0 * sqrt(N) - l) * (2.0 * sqrt(N) - l) * (2.0 * sqrt(N) - l);
    c = pow(l, 2.0 * P - 4.0);
    return (a * b * c);
  }
}

// return total wire cap
double rents(double gate_count)
{
  double gate_pitch = 2.0 * sqrt(GATE_AREA);
  double n, idf, total_cap = 0.0; 

  for (n = 1.0; n <= 2.0 * sqrt(gate_count); n ++) {
    if (n >= LOCAL_INTERMEDIATE_BOUNDARY)
      break;	 
    idf = get_idf(n, gate_count); 
    total_cap += idf * (n * gate_pitch * CW_LOCAL);
  }
  for (; n <= 2.0 * sqrt(gate_count); n ++) {
    idf = get_idf(n, gate_count); 
    total_cap += idf * (n * gate_pitch * CW_INT);
  }
  return total_cap;
}

/*****************************************/
/* temperature calculation               */
/*****************************************/

/* for debug porpuse */
void output_temperature(FILE *fp)
{
  int i;

  fprintf(fp, "\nTemperature distributions in cycle %d: \n", sim_cycle);
  //fprintf(fp, "max temp ever happening is %18.14lf\n", maxTemp);
  // note: left, bottom and right of L2 have the same temperature
  for (i = 0; i < TOTAL_MODULES; i ++) {
    fprintf(fp, "%s temperature = %18.14lf\n", module_name[i], module[i].temperature);
  }
  printf("chip_bottom temperature = %18.14lf\n", temperature_vector[TOTAL_MODULES + MODULE_L2_EXTRA + CHIP_B] - 273.15); 
  printf("spreader_west = %18.14lf\n", temperature_vector[TOTAL_MODULES + MODULE_L2_EXTRA + SP_W] - 273.15);
  printf("spreader_east = %18.14lf\n", temperature_vector[TOTAL_MODULES + MODULE_L2_EXTRA + SP_E] - 273.15);
  printf("spreader_north = %18.14lf\n", temperature_vector[TOTAL_MODULES + MODULE_L2_EXTRA + SP_N] - 273.15);
  printf("spreader_south = %18.14lf\n", temperature_vector[TOTAL_MODULES + MODULE_L2_EXTRA + SP_S] - 273.15);
  printf("spreader_bottom = %18.14lf\n", temperature_vector[TOTAL_MODULES + MODULE_L2_EXTRA + SP_B] - 273.15);
  printf("sink_west = %18.14lf\n", temperature_vector[TOTAL_MODULES + MODULE_L2_EXTRA + SINK_W] - 273.15);
  printf("sink_east = %18.14lf\n", temperature_vector[TOTAL_MODULES + MODULE_L2_EXTRA + SINK_E] - 273.15);
  printf("sink_north = %18.14lf\n", temperature_vector[TOTAL_MODULES + MODULE_L2_EXTRA + SINK_N] - 273.15);
  printf("sink_south = %18.14lf\n", temperature_vector[TOTAL_MODULES + MODULE_L2_EXTRA + SINK_S] - 273.15);
  printf("sink_bottom = %18.14lf\n", temperature_vector[TOTAL_MODULES + MODULE_L2_EXTRA + SINK_B] - 273.15);

  //fprintf(fp, "whole chip temperature = %18.14lf\n", total_tmpr);
  fprintf(fp, "End of temperature report\n\n");
}

// CACTI tool for cache or TLB partition
void cache_config_for_power(MEM_CONFIG *m, int size /* C */, int block_size /* B */, int assoc /* A */,
		  int rw_port, int w_port, int r_port)
{
  /* These variables are needed to use Cacti to auto-size cache arrays (for optimal delay) */
  result_type time_result;
  parameter_type time_parameters;
  arearesult_type arearesult;
  area_type arearesult_subbanked;
  double NSubbanks = 1.0; // assume no subbanking 
  int tag_size;

  if (assoc == 0) { // fully-associative, regonized as assoc = 0 
    // here fully-associative is only for TLB
    m->ndwl = 1;
    m->ndbl = 1;
    m->nspd = 1;
    m->ntwl = 1;
    m->ntspd = 1;

    tag_size = ADDR_WIDTH - logtwo((double)block_size);

    m->rowsb = size / block_size;
    m->colsb = tag_size;
    m->trowsb = size / block_size;
    m->tcolsb = tag_size + 1 + 6;
    m->tagSize = tag_size;
    m->totalArea = SRAM_CELL_AREA * (m->rowsb * m->colsb + m->trowsb * m->tcolsb) / U / U; 
  } else {
    time_parameters.cache_size = size; /* C */
    time_parameters.block_size = block_size; /* B */
    time_parameters.associativity = assoc; /* A */
    time_parameters.number_of_sets = size / (block_size * assoc); /* C/(B*A) */
    time_parameters.num_write_ports = w_port;
    time_parameters.num_readwrite_ports = rw_port;
    time_parameters.num_read_ports = r_port;
    time_parameters.num_single_ended_read_ports = 0;
    time_parameters.fully_assoc = 0;
    time_parameters.fudgefactor = 0.8 / TECH;
    time_parameters.tech_size = TECH;

    calculate_time(&time_result, &arearesult, &arearesult_subbanked, &time_parameters, &NSubbanks);

    m->totalArea = arearesult.totalarea;
    m->ndwl=time_result.best_Ndwl;
    m->ndbl=time_result.best_Ndbl;
    m->nspd=time_result.best_Nspd;
    m->ntwl=time_result.best_Ntwl;
    m->ntbl=time_result.best_Ntbl;
    m->ntspd=time_result.best_Ntspd;

    tag_size = ADDR_WIDTH - ((int)logtwo((double)size / block_size / assoc) + (int)logtwo((double)block_size));

    m->rowsb = size / (block_size * assoc * m->ndbl * m->nspd);
    m->colsb = 8 * block_size * assoc * m->nspd / m->ndwl;
    m->trowsb = size / (block_size * assoc * m->ntbl * m->ntspd);
    m->tcolsb = assoc * (tag_size + 1 + 6) * m->ntspd / m->ntwl;
    m->tagSize = tag_size;
  }
}

void cache_power(COMPONENT *c, MEM_CONFIG *m)
{
  c->area = m->totalArea * U * U; // total area from CACTI is in the unit of um^2
  c->inactivePower = m->ndwl * m->ndbl * inact_power_array(m->rowsb, m->colsb, c->temperature, vDD) 
  		   + m->ntwl * m->ntbl * inact_power_array(m->trowsb, m->tcolsb, c->temperature, vDD);
  c->standbyPower = m->ndwl * m->ndbl * std_power_array(m->rowsb, m->colsb, c->temperature, vDD) 
  		   + m->ntwl * m->ntbl * std_power_array(m->trowsb, m->tcolsb, c->temperature, vDD);
  c->dynPower = m->ndwl * m->ndbl * dyn_power_array(m->rowsb * m->nspd, m->colsb / m->nspd, vDD) 
  	       + m->ntwl * m->ntbl * dyn_power_array(m->trowsb * m->ntspd, m->tcolsb / m->ntspd, vDD);
  c->activePower = c->dynPower + c->standbyPower / (double)freq / U; /* convert power to uJ */
  /* activePower and dynamic power are given in cycle, while the standby power is given in the unit of Watt (not mW) */

  return;
}

// power is incremented in this function
void sram_array_power(COMPONENT *c, int row, int col, int opt)
{
  // default is for ASSIGN
  double append_factor = 0.0;
  double scale_factor = 1.0;

  if (opt == ASSIGN) {
    append_factor = 0.0;
    scale_factor = 1.0; 
  } else if (opt == APPEND) {
    append_factor = 1.0;
    scale_factor = 1.0; 
  } else {
    assert (opt > 1);
    append_factor = 0.0;
    scale_factor = (double)opt;
  }

  c->inactivePower = c->inactivePower * append_factor + scale_factor *  
		     (inact_power_array(row, col, c->temperature, vDD) + inact_power_array(row, col, c->temperature, vDD));
  c->standbyPower = c->standbyPower * append_factor + scale_factor * 
		    (std_power_array(row, col, c->temperature, vDD) + std_power_array(row, col, c->temperature, vDD));
  c->dynPower = c->dynPower * append_factor + scale_factor * (dyn_power_array(row, col, vDD) + dyn_power_array(row, col, vDD));
  c->activePower = c->dynPower + c->standbyPower / (double)freq / U; /* convert power to uJ */
  /* activePower and dynamic power are given in cycle, while the standby power is given in the unit of Watt (not mW) */

  c->area = c->area * append_factor + scale_factor * SRAM_CELL_AREA * row * col; 

  return;
}

// branch predictor power = btb + local_predict + global_predict + chooser + ras 
void branch_predictor()
{
  // BTB
  cache_power(&(module[MODULE_BRANCH]), &(mem_conf[MODULE_BRANCH]));
  // local
  sram_array_power(&(module[MODULE_BRANCH]), twolev_config[0], twolev_config[2], APPEND); // increment inside sram_array_power
  sram_array_power(&(module[MODULE_BRANCH]), twolev_config[1], 3, APPEND); // increment inside sram_array_power
  // global
  sram_array_power(&(module[MODULE_BRANCH]), bimod_config[0], 2, APPEND); // increment inside sram_array_power
  // chooser
  sram_array_power(&(module[MODULE_BRANCH]), comb_config[0], 2, APPEND); // increment inside sram_array_power
  // ras
  sram_array_power(&(module[MODULE_BRANCH]), ras_size, DATA_WIDTH, APPEND); // increment inside sram_array_power
}

// rename table power = dependence check logic + rename table
void rename_table_power()
{
  // Dependence Check Logic
  non_memory_power(&(module[MODULE_RAT]), DCL_GATE_COUNT);
  // RAT
  sram_array_power(&(module[MODULE_RAT]), MD_NUM_IREGS, (int)ceil(logtwo((double)MD_NUM_IREGS)), APPEND);
}

void non_memory_power(COMPONENT *m, unsigned int gate_count) 
{
  // Area
  m->area = gate_count * GATE_AREA;
  // Iavg to calculate standby power
  m->standbyPower = standby_power_comb(gate_count, m->temperature, vDD);
  // use CV^2 for dynamic power in uJ / cycle
  m->dynPower = dynamic_power_comb(gate_count, DYNAMIC_SWITCH_FACTOR, C_GATE, vDD); 
  // get active power
  m->activePower = m->dynPower + m->standbyPower / freq / U;
  // for the time being, assume inactive power is 1% of standby power
  m->inactivePower = 0.01 * m->standbyPower;
}

void calculate_module_power()
{
  int i;
  
  // fetch queue
  //sram_array_power(&(module[MODULE_FETCH]), ruu_ifq_size, DATA_WIDTH, ASSIGN); 

  // IALUs
  for (i = MODULE_IALU1; i <= MODULE_IALU4; i ++) { 
    //printf("ialu %d\n", i - 9);
    non_memory_power(&(module[i]), IALU_GATE_COUNT);
  }
  // FALUs
  for (i = MODULE_FALU1; i <= MODULE_FALU2; i ++) { 
    non_memory_power(&(module[i]), FALU_GATE_COUNT);
  }
 
  // rename table
  rename_table_power();

  // DECODE
  sram_array_power(&(module[MODULE_DECODE]), 65, opcode_length, ruu_issue_width);  // 65 is the total number of instructions in SimpleScalar ISA

  // INT_REG
  sram_array_power(&(module[MODULE_INTREG]), MD_NUM_IREGS, DATA_WIDTH, ASSIGN);
  // FP_REG
  sram_array_power(&(module[MODULE_FPREG]), MD_NUM_FREGS, DATA_WIDTH, ASSIGN);
 
  // RUU_SELECTION
  non_memory_power(&(module[MODULE_RUU]), RUU_SELECTION_GATE_COUNT);
  // RUU_WAKEUP
  sram_array_power(&(module[MODULE_RUU]), RUU_size, (int)ceil(logtwo((double)RUU_size)), APPEND);
  // RUU_RS
  sram_array_power(&(module[MODULE_RUU]), RUU_size, DATA_WIDTH, APPEND);
  
  // LSQ_WAKEUP // CAM is also modeled by sram array
  // FIXME: address in LSQ!
  sram_array_power(&(module[MODULE_LSQ]), LSQ_size, DATA_WIDTH, ASSIGN);
  // LSQ_RS 
  sram_array_power(&(module[MODULE_LSQ]), LSQ_size, DATA_WIDTH, APPEND);

  // FIXME
  //power->resultbus = compute_resultbus_power();

  // branch predictor power = btb + local_predict + global_predict + chooser + ras 
  // decouple btb partition and branch power calculation so that no partition require when dynamic updating power
  cache_config_for_power(&(mem_conf[MODULE_BRANCH]) /* BTB actually */, btb_config[0] * DATA_WIDTH / 8 * btb_config[1],
		DATA_WIDTH / 8, btb_config[1], 0, 1, 1); // 1 r and 1 w port 
  branch_predictor();

  // DTLB
  cache_config_for_power(&(mem_conf[MODULE_DTLB]), dtlb->nsets * itlb->bsize * 1,
	       dtlb->bsize, 0, 0, 1, 1); // 1 r and 1 w port
  cache_power(&(module[MODULE_DTLB]), &(mem_conf[MODULE_DTLB]));

  // ITLB
  cache_config_for_power(&(mem_conf[MODULE_ITLB]), itlb->nsets * itlb->bsize * 1,
	       itlb->bsize, 0, 0, 1, 1); // 1 r and 1 w port
  cache_power(&(module[MODULE_ITLB]), &(mem_conf[MODULE_ITLB]));

  // L1 icache
  cache_config_for_power(&(mem_conf[MODULE_IL1]), cache_il1->nsets * cache_il1->bsize * cache_il1->assoc,
	       cache_il1->bsize, cache_il1->assoc, 0, 1, 1); // 1 r and 1 w port
  cache_power(&(module[MODULE_IL1]), &(mem_conf[MODULE_IL1]));

  // L1 dcache
  cache_config_for_power(&(mem_conf[MODULE_DL1]), cache_dl1->nsets * cache_dl1->bsize * cache_dl1->assoc,
	       cache_dl1->bsize, cache_dl1->assoc, 2, 0, 0); // 2 r/w port
  cache_power(&(module[MODULE_DL1]), &(mem_conf[MODULE_DL1]));

  // L2 cache
  cache_config_for_power(&(mem_conf[MODULE_L2]), cache_dl2->nsets * cache_dl2->bsize * cache_dl2->assoc,
	       cache_dl2->bsize, cache_dl2->assoc, 1, 0, 0); // 1 r/w port
  cache_power(&(module[MODULE_L2]), &(mem_conf[MODULE_L2]));
}

// similar to calculate_module_power(), but no further CACTI calls
void update_module_power()
{
  int i;

  // IALUs
  for (i = MODULE_IALU1; i <= MODULE_IALU4; i ++) { 
    non_memory_power(&(module[i]), IALU_GATE_COUNT);
  }
  // FALUs
  for (i = MODULE_FALU1; i <= MODULE_FALU2; i ++) { 
    non_memory_power(&(module[i]), FALU_GATE_COUNT);
  }
  // rename table
  rename_table_power();
  // DECODE
  sram_array_power(&(module[MODULE_DECODE]), 65, opcode_length, ASSIGN); // shouldn't be the issue_width 
  // INT_REG
  sram_array_power(&(module[MODULE_INTREG]), MD_NUM_IREGS, DATA_WIDTH, ASSIGN);
  // FP_REG
  sram_array_power(&(module[MODULE_FPREG]), MD_NUM_FREGS, DATA_WIDTH, ASSIGN);
  // RUU_SELECTION
  non_memory_power(&(module[MODULE_RUU]), RUU_SELECTION_GATE_COUNT);
  // RUU_WAKEUP
  sram_array_power(&(module[MODULE_RUU]), RUU_size, (int)ceil(logtwo((double)RUU_size)), APPEND);
  // RUU_RS
  sram_array_power(&(module[MODULE_RUU]), RUU_size, DATA_WIDTH, APPEND);
  // LSQ_WAKEUP // CAM is also modeled by sram array
  // FIXME: address in LSQ!
  sram_array_power(&(module[MODULE_LSQ]), LSQ_size, DATA_WIDTH, ASSIGN);
  // LSQ_RS 
  sram_array_power(&(module[MODULE_LSQ]), LSQ_size, DATA_WIDTH, APPEND);
  // branch predictor power = btb + local_predict + global_predict + chooser + ras 
  // decouple btb partition and branch power calculation so that no partition require when dynamic updating power
  branch_predictor();
  // DTLB
  cache_power(&(module[MODULE_DTLB]), &(mem_conf[MODULE_DTLB]));
  // ITLB
  cache_power(&(module[MODULE_ITLB]), &(mem_conf[MODULE_ITLB]));
  // L1 icache
  cache_power(&(module[MODULE_IL1]), &(mem_conf[MODULE_IL1]));
  // L1 dcache
  cache_power(&(module[MODULE_DL1]), &(mem_conf[MODULE_DL1]));
  // L2 cache
  cache_power(&(module[MODULE_L2]), &(mem_conf[MODULE_L2]));
}


void output_power_temperature(tick_t sim_cycle, FILE *fp) {
	//	Total energy consumed the last time this function was called
	//	all in J
	static double total_energy_last		= 0.0;
	static double dynamic_energy_last	= 0.0;
	static double standby_energy_last	= 0.0;

	//	Total energy consumed until now
	//	all in J
	double total_energy			= 0.0;
	double dynamic_energy			= 0.0;
	double standby_energy			= 0.0;

	//	Average power consumed after the last time this function was
	//	called and until now, all in W
  	double avg_total_power_last 		= 0.0;
  	double avg_dynamic_power_last 		= 0.0; 
	double avg_standby_power_last 		= 0.0;

	//	Peak temperature over the chip surface, at this instant
	//	in deg C
  	double peak_temperature 		= 0.0;

  	int i;
  	
  	for (i = 0; i < TOTAL_MODULES; i++) {
		dynamic_energy		+= module[i].totalDynamicPower * freq * U;
		standby_energy		+= module[i].totalStandbyPower * freq * U;

		if (module[i].temperature > peak_temperature) {
			peak_temperature	= module[i].temperature;
		}
  	}  
	total_energy		= dynamic_energy + standby_energy;

	avg_total_power_last	= (total_energy - total_energy_last)/10000;
	total_energy_last	= total_energy;
	avg_dynamic_power_last	= (dynamic_energy - dynamic_energy_last)/10000;
	dynamic_energy_last	= dynamic_energy;
        avg_standby_power_last	= (standby_energy - standby_energy_last)/10000;
	standby_energy_last	= standby_energy;
	
  	fprintf(fp,"%lu\t%.3e\t%.3f\t%.3f\t%.3f\t%.3f\t%.3f\n", (unsigned long) sim_cycle, freq, vDD, avg_dynamic_power_last, avg_standby_power_last, avg_total_power_last, peak_temperature);

}   
